package syncloud.synchronization.nodes;

import syncloud.core.CoreException;
import syncloud.dao.State;
import syncloud.core.log.Logger;
import syncloud.synchronization.FolderNode;
import syncloud.synchronization.FolderContent;
import syncloud.storage.IFolder;
import syncloud.storage.INode;
import syncloud.synchronization.ISyncRecord;
import syncloud.synchronization.ISynchronizer;

import java.util.ArrayList;
import java.util.List;

public class FoldersSyncBundle extends SyncBundle {

    private static Logger logger = Logger.getLogger(FoldersSyncBundle.class);

    public FoldersSyncBundle(ISynchronizer sync, List<SyncNode> nodes) {
        super(sync, nodes);
    }

    @Override
    public void synchronize() {
        super.synchronize();
        synchronizeChildren();
    }

    public void synchronizeChildren() {
        List<IFolder> folders = new ArrayList<IFolder>();
        for(SyncNodeState syncNode: nodes) {
            INode node = syncNode.getNode();
            if (node != null) folders.add((IFolder)node);
        }
        try {
            synchronize(folders);
        } catch (CoreException e) {
            logger.error("unable to synchronize children", e);
        }
    }

    public void synchronize(List<IFolder> folders) throws CoreException {

        List<FolderContent> contents = getContents(folders);
        List<String> uniqueNames = getUniqueNames(contents);

        for(String name: uniqueNames) {
            List<FolderNode> nodes = getNodesByName(contents, name);
            List<SyncNode> nodesInfos = new ArrayList<SyncNode>();
            for (FolderNode entry: nodes) {
                SyncNode fileInfo;
                String fullName = entry.folder.getKey().child(name).getNativePath();
                fileInfo = new SyncNode(fullName, entry.folder, entry.node);
                nodesInfos.add(fileInfo);
            }
            ISyncBundle syncBundle = SyncBundleFactory.create(sync, nodesInfos);
            if (syncBundle != null) {
                syncBundle.synchronize();
            }
        }
    }

    private List<FolderContent> getContents(List<IFolder> folders) {
        List<FolderContent> contents = new ArrayList<FolderContent>();
        for(IFolder folder: folders) {
            contents.add(new FolderContent(folder, folder.getContents()));
        }
        return contents;
    }

    private List<FolderNode> getNodesByName(List<FolderContent> contents, String name) {
        List<FolderNode> nodes = new ArrayList<FolderNode>();
        for (FolderContent entry: contents) {
            INode node = find(entry.content, name);
            nodes.add(new FolderNode(entry.folder, node));
        }
        return nodes;
    }

    public List<String> getUniqueNames(List<FolderContent> contents) {
        List<String> uniqueNames = new ArrayList<String>();
        for(FolderContent folderContents: contents) {
            for(INode node: folderContents.content) {
                if (!uniqueNames.contains(node.getName()))
                    uniqueNames.add(node.getName());
            }
        }
        return uniqueNames;
    }

    private static <T extends INode> T find(List<T> items, String searchFor) {
        for(T item: items) {
            if (item.getName().equals(searchFor))
                return item;
        }
        return null;
    }

    @Override
    protected NodeState getState(SyncNode node) {
        State state = sync.getRecord().get(node.getFullname());
        if (node.getNode() == null) {
            if (state != null)
                return NodeState.DELETED;
            else
                return NodeState.NOTEXISTS;
        } else if (state != null) {
            return NodeState.UNCHANGED;
        } else {
            return NodeState.ADDED;
        }
    }

    @Override
    protected void putRecord(ISyncRecord records, INode node) {
        records.set(new State(node.getKey().getNativePath(), null));
    }

    @Override
    protected INode create(IFolder parent, INode from) throws CoreException {
        IFolder newFolder = parent.createFolder(from.getName());
        return newFolder;
    }
}
